"""
Handle conversions of MacOS installed media
"""

import os
import plistlib
import sys

from subprocess import Popen


class InstallerError(Exception):
    pass


class MacOSInstallerApp(object):
    """
    Class for MacOS installer app downloaded from app store
    """

    def __init__(self, path):
        self.path = os.path.expandvars(os.path.expanduser(path))
        self.__parse_version__()

    def __str__(self):
        return '{} (installer {})'.format(self.name, self.version)

    def __parse_version__(self):
        """
        Parse installer verison from .app version.info
        """
        version_file = os.path.join(self.path, 'Contents/Info.plist')
        if not os.path.isfile(version_file):
            raise InstallerError('No such file: {}'.format(version_file))

        try:
            data = plistlib.load(open(version_file, 'rb'))
        except Exception as e:
            raise InstallerError('Error parsing {}: {}'.format(version_file, e))

        try:
            self.version = data['CFBundleShortVersionString']
        except KeyError:
            raise InstallerError('Error looking CFBundleShortVersionString from {}'.format(version_file))

        try:
            self.volume_name = data['CFBundleDisplayName']
            self.name = data['CFBundleDisplayName'].replace('Install macOS ', '')
        except KeyError:
            raise InstallerError('Error looking CFBundleDisplayName from {}'.format(version_file))

    def convert_to_dvd(self, filename, mountpoint=None):
        """
        Convert installer to DVD image

        This is currently specific to High Sierra.

        Most likely would work with others but has not been tested.
        """

        if mountpoint is None:
            mountpoint = '/Volumes/installer_convert'

        if self.name == 'High Sierra':
            convert_command = os.path.join(
                self.path,
                'Contents/Resources/createinstallmedia'
            )
            if not os.access(convert_command, os.X_OK):
                raise InstallerError('No such executable: {}'.format(convert_command))

            # Create image file with SPUD format and HFS+J filesystem
            cmd = (
                'hdiutil',
                'create',
                '-size', '8G',
                '-layout', 'SPUD',
                '-fs', 'HFS+J',
                '-o', filename,
            )
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
            p.wait()
            # Attach image file
            cmd = (
                'hdiutil',
                'attach',
                '-noverify',
                '-mountpoint', mountpoint,
                '{}.dmg'.format(filename),
            )
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
            p.wait()

            # Use converter command to create installable DVD image. This must be run with sudo
            cmd = (
                'sudo',
                convert_command,
                '--nointeraction',
                '--applicationpath', self.path,
                '--volume', mountpoint,
            )
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
            p.wait()

            # Detach volume. After conversion it is named like the installer volume
            cmd = ('hdiutil', 'detach', '/Volumes/Install macOS High Sierra')
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
            p.wait()

            # Convert image to UDTO format
            cmd = (
                'hdiutil',
                'convert',
                '-format', 'UDTO',
                '-o', filename,
                '{}.dmg'.format(filename),
            )
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
            p.wait()

            # Rename file to remove the extra .cdr added by hdiutil
            os.rename('{}.cdr'.format(filename), filename)

            # Remove the temporary dmg file
            os.unlink('{}.dmg'.format(filename))

        else:
            raise InstallerError('Conversion of {} installer not yet supported'.format(self.name))



"""
coreStorage status classes:

just a wrapper to parse output from the command line commands to
dictionaries, grouping the hierarchy for easy consumption.
"""

from builtins import int

import re
from subprocess import check_output, CalledProcessError

re_header = re.compile('^CoreStorage logical volume groups \((\d+) found\)$')
re_lvg_header = re.compile('^Logical Volume Group ([A-Z0-9-]+)$')
re_pv_header = re.compile('^Physical Volume ([A-Z0-9-]+)$')
re_lvf_header = re.compile('^Logical Volume Family ([A-Z0-9-]+)$')
re_lv_header = re.compile('^Logical Volume ([A-Z0-9-]+)$')

re_size = re.compile('^(\d+) B \(([\d.]+) GB\)$')

LVG_HEADER_MAP = {
    'Name':                     'name',
    'Free Space':               'free_space',
    'Sequence':                 'sequence',
}

PV_HEADER_MAP = {
    'Status':                   'status',
    'Index':                    'index',
    'Disk':                     'disk',
    'Size':                     'Size',
}

LVFAMILY_HEADER_MAP = {
    'Encryption Status':        'encryption_status',
    'Encryption Context':       'encryption_content',
    'Conversion Status':        'conversion_status',
    'Sequence':                 'sequence',
    'Has Encrypted Extents':    'encrypted',
    'Conversion Direction':     'conversion_direction',
    'Encryption Type':          'encryption_type',
}

LV_HEADER_MAP = {
    'Status':               'status',
    'Sequence':             'sequence',
    'Size (Converted)':     'size_converted',
    'Size (Total)':         'size_total',
    'Volume Name':          'volume_name',
    'LV Name':              'lv_name',
    'Content Hint':         'content_hint',
    'Disk':                 'disk',
    'Revertible':           'revertible',
}


class coreStorage(list):
    """
    Class for OS/X corestorage LVM implementation status parsing
    """

    def __init__(self):
        self.lvg_count = 0
        self.update()

    def __str__(self):
        return '{0:d} groups'.format(self.lvg_count)

    def update(self):
        """
        Parse output of diskutil corestorage list to update data
        """
        lvg = None
        pv = None
        lvf = None
        lv = None

        cmd = ('diskutil', 'coreStorage', 'list')
        try:
            for line in check_output(cmd).splitlines():
                line = line.decode('utf-8').lstrip('|+-<> ')

                if line.strip() == '' or line.strip('-=') == '':
                    continue

                m = re_header.match(line)
                if m:
                    self.lvg_count = int(m.group(1))
                    continue

                m = re_lvg_header.match(line)
                if m:
                    lvg = coreStorageLVG(m.group(1))
                    self.append(lvg)
                    pv = None
                    lvf = None
                    lv = None
                    continue

                m = re_pv_header.match(line)
                if m:
                    pv = coreStoragePV(lvg, m.group(1))
                    lvg.pvs.append(pv)
                    lv = None
                    lvf = None
                    continue

                m = re_lvf_header.match(line)
                if m:
                    lvf = coreStorageLVFamily(lvg, m.group(1))
                    lvg.lvfs.append(lvf)
                    lv = None
                    continue

                m = re_lv_header.match(line)
                if m:
                    lv = coreStorageLV(lvg, m.group(1))
                    lvf.lvs.append(lv)
                    continue
                try:
                    key, value = line.strip().split(':', 1)
                except ValueError:
                    continue

                if lv is not None:
                    lv[key] = value
                elif lvf is not None:
                    lvf[key] = value
                elif pv is not None:
                    pv[key] = value
                elif lvg is not None:
                    lvg[key] = value
                else:
                    raise ValueError('Out of order line: %s' % line)

        except CalledProcessError as e:
            raise ValueError('Error listing corestorege volumes: {0}'.format(e))


class coreStorageLVG(dict):
    """
    Class to represent one corestorage LVG (logical volume group)
    """
    def __init__(self, uuid):
        self.uuid = uuid
        self.pvs = []
        self.lvfs = []

    def __str__(self):
        return 'LVG {0}\n{1}'.format(
            self.uuid,
            '\n'.join('{0:20} {1}'.format(k, v) for k, v in self.items())
        )

    def __setitem__(self, item, value):
        try:
            item = LVG_HEADER_MAP[item]
        except KeyError:
            pass
        super(coreStorageLVG, self).__setitem__(item, value)


class coreStoragePV(dict):
    """
    Class to represent one corestorage PV (physical volume)
    """
    def __init__(self, lvg, uuid):
        self.lvg = lvg
        self.uuid = uuid

    def __str__(self):
        return 'PV {0}\n{1}'.format(
            self.uuid,
            '\n'.join('{0:20} {1}'.format(k, v) for k, v in self.items())
        )

    def __setitem__(self, item, value):
        try:
            item = PV_HEADER_MAP[item]
        except KeyError:
            pass
        super(coreStoragePV, self).__setitem__(item, value)


class coreStorageLVFamily(dict):
    """
    Core storage LV (Logical Volume) family
    """
    def __init__(self, lvg, uuid):
        self.lvg = lvg
        self.uuid = uuid
        self.lvs = []

    def __str__(self):
        return 'LV Family {0}\n{1}'.format(
            self.uuid,
            '\n'.join('{0:20} {1}'.format(k, v) for k, v in self.items())
        )

    def __setitem__(self, item, value):
        try:
            item = LVFAMILY_HEADER_MAP[item]
        except KeyError:
            pass
        super(coreStorageLVFamily, self).__setitem__(item, value)


class coreStorageLV(dict):
    """
    Core storage LV (logical volume)
    """
    def __init__(self, lvf, uuid):
        self.lvf = lvf
        self.uuid = uuid

    def __str__(self):
        return 'LV {0}\n{1}'.format(
            self.uuid,
            '\n'.join('{0:20} {1}'.format(k, v) for k, v in self.items())
        )

    def __setitem__(self, item, value):
        try:
            item = LV_HEADER_MAP[item]
        except KeyError:
            pass

        if item in ('size', 'free_space', 'size_total', 'size_converted'):
            m = re_size.match(value)
            if m:
                value = int(m.group(1))

        super(coreStorageLV, self).__setitem__(item, value)



"""
OS/X implementation of the systematic VLAN management API
"""

from subprocess import Popen, PIPE

NO_VLANS_MESSAGE = 'There are no VLANs currently configured on this system.'

VLANLIST_LINE_MAP = {
    'name': 'VLAN User Defined Name:',
    'parent': 'Parent Device:',
    'port': 'Device ("Hardware" Port):',
    'tag': 'Tag:',
}


class VLANError(Exception):
    pass


class VLAN(object):
    def __init__(self):
        for k in VLANLIST_LINE_MAP.keys():
            setattr(self, k, 'NOT SET')

    def __repr__(self):
        return '{0} TAG {1} PARENT {2}'.format(self.port, self.tag, self.parent)

    def create(self, name, parent, tag):
        cmd = ['networksetup', '-createVLAN', name, parent, str(tag)]
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            raise VLANError('ERROR creating VLAN {0}'.format(tag))

        self.name = name
        self.parent = parent
        self.tag = tag

    def remove(self):
        cmd = ['networksetup', '-deleteVLAN', self.name, self.parent, str(self.tag)]
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        (stdout, stderr) = p.communicate()
        if p.returncode != 0:
            raise VLANError('ERROR removing VLAN {0}'.format(self.tag))


class VLANList(list):
    def __init__(self):
        cmd = ['networksetup', '-listVLANs']
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

        stdout, stderr = p.communicate()
        lines = [line.decode('utf-8') for line in stdout.splitlines()]
        if lines[0] == NO_VLANS_MESSAGE:
            return

        entry = None
        for line in [line.rstrip() for line in lines]:
            if line == '':
                if entry is not None:
                    self.append(entry)
                entry = None

            for k, v in VLANLIST_LINE_MAP.items():
                if line[:len(v)] == v:
                    if entry is None:
                        entry = VLAN()
                    setattr(entry, k, line[len(v):].lstrip())



"""
Module to parse output from 'ioreg' command to python data structures
"""

import os
import re
from subprocess import Popen, PIPE

IOREG_COMMAND = '/usr/sbin/ioreg'

RE_IOREG_HEADER = re.compile('^\+-o\s(?P<name>.*)\s+<class (?P<ioclass>[\w]+),(?P<flags>[^>]*)')


class IORegError(Exception):
    """
    Execptions raised by ioreg command output parsers.
    """
    pass


class IORegItem(object):
    """
    One information item line from ioreg command output
    """
    def __init__(self, line):
        try:
            key, value = line.split('=', 1)
        except ValueError:
            try:
                key, value = line.split(':', 1)
            except ValueError:
                raise IORegError('Error splitting item line {0}'.format(line))

        self.key = key.rstrip().strip('"')
        self.value = value.strip()

    def __repr__(self):
        return '{0}: {1}'.format(self.key, self.value)


class IORegGroup(dict):
    """
    A group of items in ioreg command output
    """
    def __init__(self, parent, header):
        self.parent = parent
        self.name = 'UNPARSED'

        m = RE_IOREG_HEADER.match(header)
        if not m:
            raise IORegError('Error parsing ioreg group header: {0}'.format(header))

        for k, v in m.groupdict().items():
            setattr(self, k, v.strip())

    def __repr__(self):
        return 'IORegGroup {0}'.format(self.name)

    def append(self, line):
        """
        Add an IORegItem entry to this group
        """
        line = line.strip()
        if line in ['"', '']:
            return

        item = IORegItem(line)
        self[item.key] = item


class IORegTree(list):
    """
    Parser for ioreg output entries to a dictionary
    """
    def __init__(self, path=None):
        if not os.access(IOREG_COMMAND, os.X_OK):
            raise IORegError('Not executable: {0}'.format(IOREG_COMMAND))

        if path is not None:
            cmd = [IOREG_COMMAND, '-r', '-w0', '-n', path]
        else:
            cmd = [IOREG_COMMAND, '-lw0']

        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()

        parent = None
        group = None
        for line in [l.decode('utf-8').lstrip(' |').rstrip() for l in stdout.splitlines()]:
            if line in ['', '{']:
                continue

            elif line[:3] == '+-o':
                group = IORegGroup(parent, line)
                parent = group
                self.append(group)

            elif line == '}':
                group = None

            elif group is not None:
                group.append(line)



"""
Abstraction of AppleScript system events class for python
"""

import appscript

FOLDER_NAME_MAP = {
    'applications': 'applications_folder',
    'application_support': 'application_support_folder',
    'desktop':  'desktop_folder',
    'desktop_pictures': 'desktop_pictures_folder',
    'documents': 'documents_folder',
    'downloads': 'downloads_folder',
    'favorites': 'favorites_folder',
    'fonts': 'fonts_folder',
    'home': 'home_folder',
    'library': 'library_folder',
    'movies': 'movies_folder',
    'music': 'music_folder',
    'pictures': 'pictures_folder',
    'public': 'public_folder',
    'scripts': 'scripting_additions_folder',
    'shared_documents': 'shared_documents_folder',
    'sites': 'sites_folder',
    'speakable_items': 'speakable_items_folder',
    'temp': 'temporary_items_folder',
    'trash': 'trash',
    'utilities': 'utilities_folder',
    'workflows': 'workflows_folder',
}


class SystemEventsError(Exception):
    """
    Exceptions for OS/X system events
    """
    pass


class OSXUserAccounts(dict):
    """
    List of user accounts from appscript system events
    """
    def __init__(self):
        try:
            self.app = appscript.app('System Events')
        except appscript.reference.CommandError as e:
            raise SystemEventsError('Appscript initialization error: {0}'.format(e))

        for ref in self.app.users.get():
            u = OSXUserAccount(self, ref)
            self[u.name] = u


class OSXUserAccount(dict):
    """
    One user account parsed from appscript system events API
    """
    def __init__(self, app, reference):
        self.app = app
        self.reference = reference

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            pass
        raise AttributeError

    def __getitem__(self, item):
        if item not in self.keys():
            raise KeyError('No such OSXUserAccount item: {0}'.format(item))

        if item == 'home_directory':
            return getattr(self.reference, item).get().path

        return getattr(self.reference, item).get()

    def __str__(self):
        return self.full_name

    def keys(self):
        """
        User details keys
        """
        return [k.name for k in self.reference.properties.get().keys()]

    def items(self):
        """
        User details as (key, value) list
        """
        return [(k, self[k]) for k in self.keys()]


class OSXUserFolders(dict):
    """
    List of OS/X user folders from system events API
    """
    def __init__(self):
        try:
            self.app = appscript.app('System Events')
        except appscript.reference.CommandError as e:
            raise SystemEventsError('Appscript initialization error: {0}'.format(e))

        for k in sorted(FOLDER_NAME_MAP.keys()):
            ref = getattr(self.app, FOLDER_NAME_MAP[k]).get()
            if ref is None:
                self[k] = None
                continue

            self[k] = OSXFolderItem(self, ref)


class OSXFolderItem(dict):
    """
    One OS/X folder item from system events
    """
    def __init__(self, app, reference):
        if reference is None:
            raise
        self.app = app
        self.reference = reference

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            pass
        raise AttributeError

    def __getitem__(self, item):
        if item in ['ctime', 'mtime']:
            if item == 'ctime':
                item = 'creation_date'
            if item == 'mtime':
                item = 'modification_date'
            return int(getattr(self.reference, item).get().strftime('%s'))

        if item == 'path':
            item = 'POSIX_path'

        if item not in self.keys():
            raise KeyError('No such OSXFolderItem item: {0}'.format(item))

        return getattr(self.reference, item).get()

    def __str__(self):
        return self.path

    def keys(self):
        """
        Folder detail keys
        """
        return [k.name for k in self.reference.properties.get().keys()] + ['ctime', 'mtime', 'path']

    def items(self):
        """
        Folder details as (key, value) list
        """
        return [(k, self[k]) for k in self.keys()]



"""
Wrapper to get OS/X darwin laptop battery status from command line
"""

from darwinist.ioreg import IORegTree

BATTERY_IGNORE_FIELDS = (
    'CellVoltage',
    'IOGeneralInterest',
    'LegacyBatteryInfo',
    'ManufacturerData',
)

BATTERY_FIELD_FORMATS = {
    'AdapterInfo': lambda x: int(x),
    'Amperage': lambda x: int(x),
    'AvgTimeToEmpty': lambda x: int(x),
    'AvgTimeToFull': lambda x: int(x),
    'BatteryInstalled': lambda x: x == 'Yes' and True or False,
    'BatteryInvalidWakeSeconds': lambda x: int(x),
    'BatterySerialNumber': lambda x: str(x).strip('" '),
    'CurrentCapacity': lambda x: int(x),
    'CycleCount': lambda x: int(x),
    'DesignCapacity': lambda x: int(x),
    'DeviceName': lambda x: str(x).strip('" '),
    'ExternalChargeCapable': lambda x: x == 'Yes' and True or False,
    'ExternalConnected': lambda x: x == 'Yes' and True or False,
    'FirmwareSerialNumber': lambda x: int(x),
    'FullyCharged': lambda x: x == 'Yes' and True or False,
    'InstantAmperage': lambda x: int(x),
    'InstantTimeToEmpty': lambda x: int(x),
    'IsCharging': lambda x: x == 'Yes' and True or False,
    'Location': lambda x: int(x),
    'Manufacturer': lambda x: str(x).strip('" '),
    'MaxCapacity': lambda x: int(x),
    'MaxErr': lambda x: int(x),
    'PermanentFailureStatus': lambda x: int(x),
    'PostChargeWaitSeconds': lambda x: int(x),
    'PostDischargeWaitSeconds': lambda x: int(x),
    'Temperature': lambda x: float(x) / 100,
    'TimeRemaining': lambda x: int(x),
    'Voltage': lambda x: float(x) / 1000,
}


class Batteries(list):
    """
    All connected OS/X computer batteries based on ioreg data
    """
    def __init__(self):
        ioreg_data = IORegTree('AppleSmartBattery')
        for ioreg_group in ioreg_data:
            self.append(Battery(ioreg_group))


class Battery(dict):
    """Battery details

    """
    def __init__(self, details=None):
        if details:
            for key, value in details.items():
                if key in BATTERY_IGNORE_FIELDS:
                    continue

                if key in BATTERY_FIELD_FORMATS.keys():
                    self[key.lower()] = BATTERY_FIELD_FORMATS[key](value.value)
                else:
                    self[key.lower()] = value.value

        if 'currentcapacity' in self and 'maxcapacity' in self:
            self['percent'] = int(round(float(self.currentcapacity) / self.maxcapacity * 100))
        else:
            self['percent'] = 'UNKNOWN'

    def __getattr__(self, attr):
        try:
            return self[attr.lower()]
        except KeyError:
            raise AttributeError('No such Battery attribute: {0}'.format(attr))

    def __repr__(self):
        if self.ischarging:
            prefix = 'CHARGING'
        elif self.fullycharged:
            prefix = 'FULL'
        else:
            prefix = 'DISCHARGING'
        return '{0} {1} {2:d}% {3:d}/{4:d} mAh {5:d} cycles'.format(
            self.devicename,
            prefix,
            self.percent,
            self.currentcapacity,
            self.maxcapacity,
            self.cyclecount,
        )

    def keys(self):
        return sorted(dict.keys(self))

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        return [self[k] for k in self.keys()]



"""
Send notifications to OS/X Mountain Lion notification center
"""

import Foundation
import objc
import time


class NotificationCenter(object):
    def __init__(self):
        self.notification_center = objc.lookUpClass('NSUserNotificationCenter')

    @property
    def user_notification_center(self):
        return

    def notify(self, title, subtitle, text, url=None):
        client = Notification.alloc().init()

        user_notification_center = self.notification_center.defaultUserNotificationCenter()
        user_notification_center.setDelegate_(client)

        notification = objc.lookUpClass('NSUserNotification').alloc().init()
        notification.setTitle_(str(title))
        notification.setSubtitle_(str(subtitle))
        notification.setInformativeText_(str(text))
        notification.setSoundName_('NSUserNotificationDefaultSoundName')

        if url is not None:
            notification.setUserInfo_({'action': 'open_url', 'value': url})

        user_notification_center.scheduleNotification_(notification)

        time.sleep(5)
        return client


class Notification(Foundation.NSObject):
    def userNotificationCenter_didDeliverNotification_(self, center, notification):
        pass

    def userNotificationCenter_didActivateNotification_(self, center, notification):
        pass

        userInfo = notification.userInfo()
        if userInfo['action'] == 'open_url':
            import subprocess
            subprocess.Popen(['open', userInfo['value']])



"""
Wrapper classes to store configuration file of dmg parameters and to attach
and detach the configured DMGs.
"""

import os
import sys

from subprocess import Popen, PIPE
from configobj import ConfigObj, Section

from systematic.shell import CONFIG_PATH
from darwinist.diskutil import DiskInfo, DiskUtilError

DEFAULT_CONFIG_PATH = os.path.join(CONFIG_PATH, 'diskimages.conf')
DEFAULT_VOLUMES = '/Volumes/'
VALID_CONFIG_ARGS = (
    'description',
    'image',
    'mountpoint',
    'args',
)


class DiskImageError(Exception):
    pass


class DiskImagesConfig(dict):
    def __init__(self, path=DEFAULT_CONFIG_PATH):
        self.path = path
        if os.path.isfile(path):
            self.read()

    def read(self):
        config = ConfigObj(self.path)

        for name, options in config.items():
            if not isinstance(options, Section):
                continue

            for arg in options.keys():
                if arg not in VALID_CONFIG_ARGS:
                    raise DiskImageError('Invalid option: {0}'.format(arg))

            self[name] = DiskImage(self, name, **options)

    def keys(self):
        return sorted(super(DiskImagesConfig, self).keys())

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        return [self[k] for k in self.keys()]

    def match(self, value):
        if value in self.keys():
            return self[value]

        for dmg in self.values():
            if dmg.image == value:
                return dmg

            if dmg.mountpoint == value:
                return dmg

        return None


class DiskImage(object):
    def __init__(self, config, name, image, mountpoint, args, description=None):
        self.config = config
        self.name = name
        self.image = image
        self.mountpoint = mountpoint
        self.description = description is not None and description or ''

        if not isinstance(args, list):
            args = [args]

        if not self.mountpoint[:len(DEFAULT_VOLUMES)] == DEFAULT_VOLUMES:
            args.extend(['-mountpoint', self.mountpoint])

        self.args = args

    def __repr__(self):
        return '{0} mounted on {1} ({2})'.format(
            self.image,
            self.mountpoint,
            ' '.join(self.args),
        )

    def __getattr__(self, attr):
        if attr == 'connected':
            try:
                if 'MountPoint' in self.info:
                    return True
                return False
            except AttributeError:
                return False

        if attr == 'info':
            try:
                return DiskInfo(self.mountpoint)
            except DiskUtilError:
                return {}

        raise AttributeError('No such DiskImage attribute: {0}'.format(attr))

    def detach(self):
        if not self.connected:
            raise DiskImageError('Not attached: {0}'.format(self.mountpoint))

        cmd = ('hdiutil', 'detach', self.mountpoint)
        p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
        p.wait()
        if p.returncode != 0:
            raise DiskImageError('Error detaching {0}'.format(self.mountpoint))

    def attach(self, passphrase=None):
        if self.connected:
            raise DiskImageError('Already attached: {0}'.format(self.mountpoint))

        cmd = ['hdiutil', 'attach'] + self.args + [self.image]

        if passphrase is None:
            p = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
        else:
            p = Popen(cmd, stdin=PIPE, stdout=sys.stdout, stderr=sys.stderr)
            stdout, stderr = p.communicate(passphrase)

        p.wait()
        if p.returncode != 0:
            raise DiskImageError('Error attaching {0}'.format(self.image))



"""
Wrapper to call the networksetup command line script.
"""

import os
import re
from subprocess import check_output, CalledProcessError
from seine.address import IPv4Address, EthernetMACAddress

CMD = '/usr/sbin/networksetup'

RE_ETHERNET_ADDRESS = re.compile(
    'Ethernet Address: (?P<mac>.*) \(Hardware Port: (?P<port>[^/)]+)\)$'
)


class NetworkSetupError(Exception):
    """
    Exceptions raised while parsing OS/X network setup
    """
    pass


class NetworkSetup(object):
    """
    Wrapper for OS/X 'networksetup' command line tool
    """
    def __init__(self):
        if not os.path.isfile(CMD):
            raise NetworkSetupError('No such file: {0}'.format(CMD))
        if not os.access(CMD, os.X_OK):
            raise NetworkSetupError('No permission to execute: {0}'.format(CMD))

        self.network_services = []
        for line in check_output(('networksetup', '-listallnetworkservices')).splitlines():
            line = line.decode('utf-8')
            if line != '':
                self.network_services.append(NetworkService(line))

    def __getattr__(self, attr):
        raise AttributeError('No such NetworkSetup attribute: {0}'.format(attr))


class NetworkService(object):
    """
    One specific network service parsed from 'networksetup' output
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    def __parse_info(self):
        """
        Parser for the device details (called when attributes are requested)
        """
        details = {}
        try:
            out = filter(lambda x: x != '', check_output(
                ['networksetup', '-getinfo', self.name]
            ).splitlines())
        except CalledProcessError:
            return details

        for line in out:
            if line == 'Manual Configuration':
                details['ipv4_mode'] = 'manual'
                continue

            if line == 'DHCP Configuration':
                details['ipv4_mode'] = 'dhcp'
                continue

            try:
                key, value = [x.strip() for x in line.split(':', 1)]
            except ValueError:
                raise ValueError('Error parsing line: {0}'.format(line))

            if value == 'none':
                value = None

            if key in ['Subnet mask', 'IP address', 'Router']:
                value = IPv4Address(value)

            if key == 'Ethernet Address':
                value = EthernetMACAddress(value)

            details[key] = value

        return details

    def __parse_mac_address(self):
        """
        Parser for device MAC address
        """
        try:
            out = check_output(
                ['networksetup', '-getmacaddress', self.name]
            ).strip('\n')
        except CalledProcessError:
            return None
        m = RE_ETHERNET_ADDRESS.match(out)
        if not m:
            return None
        mac = m.groupdict()['mac']
        if mac == '(null)':
            return None
        return mac

    @property
    def configuration(self):
        return self.__parse_info()

    @property
    def mac_address(self):
        return self.__parse_mac_address()

    def set_mode(self, mode, client_id=None, ipaddress=None, netmask=None, router=None):
        """
        Set interface mode to given value. Each mode has it's specific allowed flags,
        see OS/X documentation or source code of this module.

        Mode can be: dhcp, manual or bootp
        """
        if mode not in ['dhcp', 'manual', 'bootp']:
            raise ValueError('Invalid IPv4 configuratio mode')

        if mode == 'bootp':
            try:
                check_output(['networksetup', '-setbootp', self.name])
            except CalledProcessError:
                raise NetworkSetupError('Error setting {0} to BOOTP mode'.format(self.name))

        elif mode == 'dhcp':
            cmd = ['networksetup', '-setdhcp', self.name]
            if client_id is not None:
                cmd.append(client_id)
            try:
                check_output(cmd)
            except CalledProcessError:
                raise NetworkSetupError('Error setting {0} to DHCP mode'.format(self.name))

        elif mode == 'manual':
            if ipaddress is None:
                raise NetworkSetupError('Manual mode requires valid ipaddress')
            else:
                try:
                    ipaddress = IPv4Address(ipaddress)
                except ValueError:
                    raise NetworkSetupError('Invalid IPv4 Address: {0}'.format(ipaddress))

            if netmask is None:
                raise NetworkSetupError('Manual mode requires valid netmask')
            else:
                try:
                    netmask = IPv4Address(netmask)
                except ValueError:
                    raise NetworkSetupError('Invalid IPv4 netmask: {0}'.format(netmask))

            if router is None:
                raise NetworkSetupError('Manual mode requires valid router')
            else:
                try:
                    router = IPv4Address(router)
                except ValueError:
                    raise NetworkSetupError('Invalid IPv4 Address: {0}'.format(router))

            cmd = [
                'networksetup',
                '-setmanual', self.name,
                ipaddress.ipaddress,
                netmask.ipaddress,
                router.ipaddress
            ]

            try:
                return check_output(cmd)
            except CalledProcessError:
                raise NetworkSetupError('Error setting {0} to manual mode'.format(self.name))



"""
Control and see status of OS/X network interfaces from python.
"""

import os
import appscript
from configobj import ConfigObj

NETWORK_CONNECTION_TYPES = {
    'wlan':         [2],
    'thunderbolt':  [5],
    'ethernet':     [6],
    'dialup':       [7, 8],  # Both bluetooth and USB 3G connections
    'firewire':     [9],
    'vpn':          [10, 13],
}


class NetworkConfigError(Exception):
    pass


class NetworkProfileList(object):
    """
    List of network profiles found on the OS/X system
    """
    def __init__(self, config=None):
        self.config = self.__read_config__(config)

        try:
            self.app = appscript.app('System Events')
            self.location = self.app.network_preferences.get().current_location.get()

        except appscript.reference.CommandError as e:
            raise NetworkConfigError('Appscript initialization error: {0}'.format(e))

    def __read_config__(self, config):
        if not os.path.isfile(config):
            return {}
        return ConfigObj(config)

    def __getitem__(self, item):
        names = filter(
            lambda s: s.name.get() == item,
            self.location.services.get()
        )
        if not names:
            raise KeyError('No such connection: {0}'.format(item))

        try:
            return NetworkConnection(item, profiles=self)

        except appscript.reference.CommandError:
            raise KeyError('No such network service configured: {0}'.format(item))

    def keys(self):
        """
        Return names of network locations
        """
        return [s.name.get() for s in self.location.services.get()]

    def items(self):
        """
        Return (name,NetworkConnection() list based on self.keys()
        """
        return [(k, NetworkConnection(k, profiles=self)) for k in self.keys()]

    def values(self):
        """
        Return NetworkConnection() list based on self.keys()
        """
        return [NetworkConnection(k, profiles=self) for k in self.keys()]

    def filter(self, connection_type):
        """
        Return network connections matching connection_type
        """
        try:
            connection_types = int(connection_type)

        except ValueError:
            try:
                connection_types = NETWORK_CONNECTION_TYPES[connection_type]
            except KeyError:
                raise NetworkConfigError('Unknown connection type: {0}'.format(connection_type))

        if not isinstance(connection_types, list):
            connection_types = [connection_types]

        filtered = []
        for service in self.location.services.get():
            if service.kind.get() in connection_types:
                filtered.append(NetworkConnection(service.name.get(), self))

        return filtered


class NetworkConnection(object):
    """
    Details and control of one OS/X network connection
    """
    def __init__(self, name, profiles=None):
        if profiles is None:
            profiles = NetworkProfileList()
        self.name = name
        self.app = profiles.app

        names = filter(lambda s: s.name.get() == name, profiles.location.services.get())
        if not len(names):
            raise NetworkConfigError('No such connection: {0}'.format(name))

        try:
            self.connection = profiles.location.services[name].get()
        except KeyError as e:
            raise NetworkConfigError(str(e).strip("'"))

    def __eq__(self, other):
        if not hasattr(other, 'app'):
            return 0

        if self.connection_type != other.connection_type:
            return

    def __hash__(self):
        return '{0} {1}'.format(self.connection_type, self.name)

    def __repr__(self):
        return '{0} {1}: {2}'.format(
            self.connection_type,
            self.name,
            self.connected and 'connected' or 'not connected'
        )

    def get_interface_attribute(self, item):
        try:
            interface = self.connection.interface.get()
            return getattr(interface, item).get()

        except appscript.reference.CommandError:
            raise AttributeError('Value {0} not available for interface {1}'.format(
                item,
                self.name
            ))

    @property
    def mac(self):
        return self.get_interface_attribute('MAC_address')

    @property
    def mac_address(self):
        return self.get_interface_attribute('MAC_address')

    @property
    def kind(self):
        try:
            return self.connection.kind.get()

        except appscript.reference.CommandError:
            raise NetworkConfigError('Error getting service type for {0}'.format(self.name))

    @property
    def connection_type(self):
        for k, values in NETWORK_CONNECTION_TYPES.items():
            if self.kind in values:
                return k
        return 'unknown'

    @property
    def connected(self):
        try:
            return self.connection.current_configuration.connected.get()

        except appscript.reference.CommandError:
            return False

    def connect(self):
        """
        Connects the network interface.

        If the interface is a VPN and it requires a password, you will still
        get graphical password request dialog. Patches welcome how to enter
        the passphrase with appscript.
        """
        if self.connected:
            raise NetworkConfigError('Already connected: {0}'.format(self.name))

        self.app.connect(self.connection)

    def disconnect(self):
        """
        Disconnect the network interface
        """
        if not self.connected:
            raise NetworkConfigError('Not connected: {0}'.format(self.name))

        self.app.disconnect(self.connection)



"""
OS/X application bundles
"""

import os
import plistlib

from xml.parsers.expat import ExpatError

INFO_BUNDLE_NAME_MAP = {
    'name':             'CFBundleName',
    'bundle_id':        'CFBundleIdentifier',
    'version':          'CFBundleVersion',
    'short_version':    'CFBundleShortVersionString',
}


class ApplicationError(Exception):
    pass


class Application(object):
    """
    Class to parse OS/X application bundle (.app) directory
    """

    def __init__(self, path):
        self.path = path
        self.__cached_info = None

        if not os.path.isdir(path):
            raise ApplicationError('No such directory: {0}'.format(self.path))

        if os.path.splitext(os.path.realpath(path))[1] != '.app':
            raise ApplicationError('Not an application bundle: {0}'.format(path))

    def __repr__(self):
        return self.path

    @property
    def info(self):
        if not self.__cached_info:
            self.__cached_info = ApplicationInfo(self)
        return self.__cached_info

    @property
    def version(self):
        info = self.info

        for key in ('version', 'short_version'):
            value = getattr(info, key)
            if value is not None:
                return value

        return 'UNKNOWN'


class ApplicationInfo(dict):
    """
    Information for an application, as dictionary
    """

    def __init__(self, application):
        self.path = os.path.join(application.path, 'Contents', 'Info.plist')

        if not os.path.isfile(self.path):
            raise ApplicationError('No such file: {0}'.format(self.path))

        try:
            self.update(plistlib.readPlist(self.path).items())
        except ExpatError as e:
            raise ApplicationError('Error parsing {0}: {1}'.format(self.path, e))

    def __repr__(self):
        return str('{0} {1}'.format(self.name, self.version))

    def __getattr__(self, attr):
        if attr in INFO_BUNDLE_NAME_MAP.keys():
            attr = INFO_BUNDLE_NAME_MAP[attr]

        try:
            return self[attr]
        except KeyError:
            return None


class ApplicationTree(list):
    """
    Tree of OS/X applications (.app directory bundles)
    """

    def __init__(self, path='/Applications', max_depth=2):
        self.path = path
        self.max_depth = max_depth
        self.update()

    def update(self):
        """
        Update application tree recursively with load_tree()
        """

        def load_tree(path, depth=0):
            """
            Load tree items from given path
            """
            if depth >= self.max_depth:
                return []

            apps = []
            for s in sorted(os.path.join(path, d) for d in os.listdir(path)):
                if not os.path.isdir(s):
                    continue

                if os.path.splitext(s)[1][1:] == 'app':
                    apps.append(Application(s))
                else:
                    apps.extend(load_tree(s, depth=depth+1))

            return apps

        if not os.path.isdir(self.path):
            return

        self.__delslice__(0, len(self))
        self.extend(load_tree(self.path))



"""
Wrapper for OS/X diskutil command for python
"""

import os
import plistlib
from xml.parsers.expat import ExpatError
from subprocess import Popen, PIPE

from io import BytesIO

INFO_FIELD_MAP = {
    'DeviceNode':       {'name': 'Device', 'value': lambda x: str(x)},
    'FilesystemName':   {'name': 'Filesystem', 'value': lambda x: str(x)},
    'UsedSpace':        {'name': 'Used', 'value': lambda x: x/1024},
    'UsedPercent':      {'name': 'Percent', 'value': lambda x: x/1024},
    'FreeSpace':        {'name': 'Free', 'value': lambda x: x/1024},
    'TotalSize':        {'name': 'Sizd', 'value': lambda x: x/1024},
    'VolumeName':       {'name': 'Volume Name', 'value': lambda x: str(x)},
    'VolumeUUID':       {'name': 'UUID', 'value': lambda x: str(x)},
}
INFO_FIELD_ORDER = [
    'DeviceNode',
    'VolumeName',
    'FilesystemName',
    'VolumeUUID',
    'UsedSpace',
    'FreeSpace',
    'TotalSize'
]


class DiskUtilError(Exception):
    pass


class DiskInfo(dict):
    def __init__(self, device):
        if not os.access(device, os.R_OK):
            raise DiskUtilError('Device not readable: {0}'.format(device))

        cmd = ('diskutil', 'info', '-plist', device)
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        try:
            plist = BytesIO(stdout)
            self.update(plistlib.readPlist(plist))
        except ExpatError:
            raise DiskUtilError('Error parsing plist: {0}'.format(stdout))

        if 'TotalSize' in self and 'FreeSpace' in self:
            self['UsedSpace'] = self.TotalSize - self.FreeSpace
            self['UsedPercent'] = int(round(1-(float(self.FreeSpace) / float(self.TotalSize))))

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError

    def keys(self):
        """
        Return keys as sorted list
        """
        return sorted(super(DiskInfo, self).keys())

    def items(self):
        """
        Return (key, value) sorted by key
        """
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        """
        Return values sorted by key
        """
        return [self[k] for k in self.keys()]



"""
IOS mobile backups parser
"""

import os
import plistlib
import sqlite3

from operator import attrgetter
from datetime import datetime, timedelta
from subprocess import Popen, PIPE
from xml.parsers.expat import ExpatError

# Path of mobile backups on OS/X
BACKUP_PATH = os.path.expanduser('~/Library/Application Support/MobileSync/Backup')

# Database hashed names in backups
DATABASE_FILENAMES = {
    'sms':      '3d0d7e5fb2ce288813306e4d4636395e047a3d28',
    'contacts': '31bb7ba8914766d4ba40d6dfb6113c8b614be442',
    'calendar': '2041457d5fe04d39d0ab481178355df6781e6858',
    'notes':    'ca3bc056d4da0bbf88b5fb3be254f3b7147e639c',
    'calls':    '2b2b0084a1bc3a5ac8c27afdf14afb42c61a19ca',
}

# Message dates are stored starting 2001-01-01
START_DATE = datetime(year=2001, month=1, day=1, hour=0, minute=0, second=0)

# Standard labels in addressbok contacts
CONTACTS_LABEL_MAP = {
    1:  'telephone',
    2:  'work',
    3:  'homepage',
    4:  'mobile',
    5:  'home',
}


class IOSBackupError(Exception):
    pass


class IOSDatabaseBackup(object):
    name = 'no name'
    db_hash = ''

    def __init__(self, backup):
        self.backup = backup
        self.path = os.path.join(backup.path, self.db_hash)
        self.__connection__ = None
        self.__cached_data__ = {}

    @property
    def exists(self):
        return os.path.isfile(self.path)

    @property
    def readable(self):
        return os.access(self.path, os.R_OK)

    @property
    def connection(self):
        if self.__connection__ is not None:
            return self.__connection__
        self.__connection__ = sqlite3.Connection(self.path)
        return self.__connection__

    @property
    def cursor(self):
        return self.connection.cursor()


class SortedContainer(object):

    sort_keys = ()

    def __init__(self):
        self.__cached_data__ = {}


class Handle(SortedContainer):
    sort_keys = ('country', 'service', 'id', 'number')

    def __init__(self, database, handle_id, country, service, number):
        super(Handle, self).__init__()

        self.id = int(handle_id)
        self.country = country
        self.service = service
        self.number = number

    def __repr__(self):
        return self.number


class Message(SortedContainer):
    sort_keys = ('database', 'date')

    def __init__(self, database, message_id, sender_handle_id, date, subject, text, is_from_me):
        super(Message, self).__init__()

        self.database = database
        self.id = message_id
        self.subject = subject
        self.text = text
        self.is_from_me = is_from_me == 1

        self.handle = self.database.find_handle(sender_handle_id)
        self.date = START_DATE + timedelta(seconds=date)

    @property
    def sender(self):
        if self.is_from_me:
            return 'ME'
        else:
            contact = self.database.backup.addressbook.lookup_by_number(self.handle.number)
            if contact is not None:
                return contact
            else:
                return 'UNKNOWN'


class Chat(SortedContainer):

    sort_keys = ('id',)

    def __init__(self, database, chat_id):
        self.database = database
        self.id = chat_id

    @property
    def first(self):
        try:
            return self.messages[0]
        except IndexError:
            return None

    @property
    def latest(self):
        try:
            return self.messages[-1]
        except IndexError:
            return None

    @property
    def messages(self):
        cursor = self.database.cursor
        cursor.execute("""SELECT message_id FROM chat_message_join WHERE chat_id=?""", (self.id,))
        message_ids = [v[0] for v in cursor.fetchall()]
        messages = []
        for message in self.database.messages:
            if message.id in message_ids:
                messages.append(message)
        return sorted(messages, key=attrgetter(*Message.sort_keys))


class SMSDatabase(IOSDatabaseBackup):
    name = 'sms'
    db_hash = '3d0d7e5fb2ce288813306e4d4636395e047a3d28'

    @property
    def handles(self):
        try:
            return self.__cached_data__['handles']
        except KeyError:
            return self.fetch_handles()

    @property
    def messages(self):
        try:
            return self.__cached_data__['messages']
        except KeyError:
            return self.fetch_messages()

    @property
    def chats(self):
        try:
            return self.__cached_data__['chats']
        except KeyError:
            return self.fetch_chats()

    def fetch_handles(self):
        cursor = self.cursor
        cursor.execute("""SELECT rowid, country, service, id FROM handle""")
        self.__cached_data__['handles'] = sorted(
            [Handle(self, *data) for data in cursor.fetchall()],
            key=attrgetter(Handle.sort_keys)
        )
        return self.__cached_data__['handles']

    def fetch_messages(self):
        cursor = self.cursor
        cursor.execute("""
            SELECT rowid AS message_id, handle_id AS sender_handle_id, date, subject, text, is_from_me
            FROM message
        """)
        self.__cached_data__['messages'] = sorted(
            [Message(self, *data) for data in cursor.fetchall()],
            key=attrgetter(*Message.sort_keys)
        )
        return self.__cached_data__['messages']

    def fetch_chats(self):
        cursor = self.cursor
        cursor.execute("""SELECT rowid FROM chat""")
        self.__cached_data__['chats'] = sorted(
            [Chat(self, *data) for data in cursor.fetchall()],
            key=attrgetter(*Chat.sort_keys)
        )
        return self.__cached_data__['chats']

    def find_handle(self, handle_id):
        for handle in self.handles:
            if handle.id == handle_id:
                return handle
        return None


class ContactProperty(object):
    def __init__(self, contact, label, value):
        self.contact = contact
        self.label = label
        self.value = value

    def __repr__(self):
        return '%s %s' % (self.name, self.value)

    @property
    def name(self):
        try:
            return CONTACTS_LABEL_MAP[self.label]
        except KeyError:
            return self.label


class Contact(SortedContainer):
    sort_keys = ('last', 'middle', 'first')

    def __init__(self, database, contact_id, first, last, middle):
        super(Contact, self).__init__()

        self.database = database
        self.id = contact_id
        self.first = first is not None and first or ''
        self.last = last is not None and last or ''
        self.middle = middle is not None and middle or ''

        self.__cached_data = {}

    def __repr__(self):
        name = ''
        for value in (self.first, self.middle, self.last):
            if value:
                name += '%s ' % value
        return name.strip()

    @property
    def properties(self):
        try:
            return self.__cached_data__['properties']
        except KeyError:
            return self.fetch_properties()

    def fetch_properties(self):
        cursor = self.database.cursor
        cursor.execute("""SELECT label, value FROM ABMultiValue WHERE record_id=?""", (self.id, ))
        self.__cached_data__['properties'] = [ContactProperty(self, *data) for data in cursor.fetchall()]
        return self.__cached_data__['properties']


class AddressbookDatabase(IOSDatabaseBackup):
    name = 'contacts'
    db_hash = '31bb7ba8914766d4ba40d6dfb6113c8b614be442'

    def __init__(self, backup):
        super(AddressbookDatabase, self).__init__(backup)

    @property
    def contacts(self):
        try:
            return self.__cached_data__['contacts']
        except KeyError:
            return self.fetch_contacts()

    def fetch_contacts(self):
        cursor = self.cursor
        cursor.execute("""SELECT rowid AS contact_id, first, last, middle FROM ABPerson""")
        self.__cached_data__['contacts'] = sorted(Contact(self, *data) for data in cursor.fetchall())
        return self.__cached_data__['contacts']

    def lookup_by_number(self, number):
        for contact in self.contacts:
            for prop in contact.properties:
                if prop.value is None:
                    continue
                if number == prop.value.replace(' ', ''):
                    return contact
        return None


class CalendarDatabase(IOSDatabaseBackup):
    name = 'calendar'
    db_hash = '2041457d5fe04d39d0ab481178355df6781e6858'

    def __init__(self, backup):
        IOSDatabaseBackup.__init__(self, backup)


class NotesDatabase(IOSDatabaseBackup):
    name = 'notes'
    db_hash = 'ca3bc056d4da0bbf88b5fb3be254f3b7147e639c'

    def __init__(self, backup):
        IOSDatabaseBackup.__init__(self, backup)


class CallsDatabase(IOSDatabaseBackup):
    name = 'calls'
    db_hash = '2b2b0084a1bc3a5ac8c27afdf14afb42c61a19ca'

    def __init__(self, backup):
        IOSDatabaseBackup.__init__(self, backup)


class IOSBackup(object):
    def __init__(self, path):
        self.path = path

        self.sms = SMSDatabase(self)
        self.addressbook = AddressbookDatabase(self)
        self.notes = NotesDatabase(self)
        self.calendar = CalendarDatabase(self)
        self.calls = CallsDatabase(self)

        self.configuration = '248ed6d5d0a8c3a9cc5f8bd2048aac03b273f296'

    def __repr__(self):
        return '%s (updated %s)' % (self.device_name, self.updated)

    def __read_binary_plist__(self, path):
        p = Popen(['plutil', '-convert', 'xml1', '-o', '-', path], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        return plistlib.readPlistFromString(stdout)

    @property
    def device_name(self):
        path = os.path.join(self.path, '13fcec800c483aa9cc21b0f0e731757ac0f2dea9')
        if not os.path.isfile(path):
            raise IOSBackupError('No such file: {0}'.format(path))

        try:
            plist = self.__read_binary_plist__(path)
        except ExpatError as e:
            raise IOSBackupError('Error reading {0}: {1}'.format(path, e))

        try:
            return plist['UserAssignedDeviceName']
        except KeyError:
            return None

    @property
    def updated(self):
        try:
            return datetime.fromtimestamp(os.stat(self.path).st_mtime)
        except OSError as e:
            raise IOSBackupError('Error checking mtime of %s: %s' % (self.path, e))

    @property
    def id(self):
        return os.path.basename(self.path)


class IOSDeviceBackups(list):
    def __init__(self, path=BACKUP_PATH):
        if not os.path.isdir(path):
            raise IOSBackupError('Not a directory: {}'.format(path))

        try:
            paths = [os.path.join(path, name) for name in os.listdir(path)]
        except OSError as e:
            raise IOSBackupError('Error listing directory {}: {}'.format(path, e))

        for path in paths:
            self.append(IOSBackup(path))



"""
OSX core location API for python
"""

import json
import requests
import time
import CoreLocation

API_URL = """http://maps.googleapis.com/maps/api/geocode/json?latlng=%(latitude)s,%(longitude)s&sensor=false"""


class CoreLocationError(Exception):
    pass


class CoreLocationManager(object):
    def __init__(self):
        self.manager = CoreLocation.CLLocationManager.alloc().init()
        self.startUpdatingLocation()

    def stopUpdatingLocation(self):
        self.manager.stopUpdatingLocation()

    def startUpdatingLocation(self):
        self.manager.startUpdatingLocation()

    def get_coordinates(self, seconds=5):
        """Get location

        Returns coordinates from core location API
        """
        wait = seconds
        while True:
            location = self.manager.location()
            if location is not None:
                return location.coordinate()

            wait -= 1
            if wait == 0:
                break
            time.sleep(1)

        raise CoreLocationError('Error getting location in {0} seconds'.format(seconds))

    def get_addresses(self, coordinates, count=1):
        url = API_URL % {'latitude': coordinates.latitude, 'longitude': coordinates.longitude, }
        response = requests.get(url)
        if response.status_code != 200:
            raise CoreLocationError('Error querying google maps API: returns status code {0}'.format(
                response.status_code
            ))

        addresses = []
        try:
            data = json.loads(response.content)
            for record in data['results']:
                if 'formatted_address' in record:
                    address = record['formatted_address']
                    if address not in addresses:
                        addresses.append(address)

                    if len(addresses) >= count:
                        break

        except ValueError:
            raise CoreLocationError('Error parsing google maps location API response as JSON')

        except KeyError:
            raise CoreLocationError('Could not find formatted_address from JSON response')

        return addresses



"""
Timemachine control.

Wraps tmutil command for use from python
"""

import os
import re

from datetime import datetime
from subprocess import Popen, PIPE

TMUTIL = '/usr/bin/tmutil'
TMUTIL_VERSION_PATTERNS = (
    re.compile('^tmutil version (?P<version>[^ ]+) \(built (?P<builddate>[^)]+)\)$'),
)
TMUTIL_VERSION_DATE_PATTERNS = (
    '%b %d %Y',
)


class TimeMachineError(Exception):
    pass


class TmUtilVersion(dict):
    """Parser for tmutil version

    """
    def __init__(self, value):
        for pattern in TMUTIL_VERSION_PATTERNS:
            m = pattern.match(value)
            if m:
                self.update(m.groupdict())
                if 'builddate' in self:
                    for datefmt in TMUTIL_VERSION_DATE_PATTERNS:
                        try:
                            self['builddate'] = datetime.strptime(self['builddate'], datefmt)
                            break
                        except:  # noqa
                            pass
                    if not isinstance(self['builddate'], datetime):
                        raise TimeMachineError('Error parsing build date {0}'.format(self['builddate']))
                return

        raise TimeMachineError('Error parsing tmutil version from {0}'.format(value))


class TimeMachineDestination(dict):
    """Time machine destination

    """
    def __init__(self, data):
        for line in [line.decode('utf-8').rstrip() for line in data.splitlines() if not line.startswith('=')]:
            try:
                key, value = [x.strip() for x in line.split(':', 1)]
                self[key] = value
            except Exception as e:
                raise TimeMachineError('Error parsing destination info line {0}: {1}'.format(line, e))


class TimeMachineUtility(object):
    """Wrap tmutil cli

    Note: not all args are parsed yet.
    """

    def __init__(self):
        if not os.access(TMUTIL, os.X_OK):
            raise TimeMachineError('Not executable: {0}'.format(TMUTIL))

    def __run__(self, args):
        """Run tmutil command

        """
        if isinstance(args, str):
            args = [args]

        cmd = [TMUTIL] + args
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        return p.returncode, stdout, stderr

    @property
    def version(self):
        """Tmutil version

        """
        rv, stdout, stderr = self.__run__('version')
        return TmUtilVersion(stdout.strip())

    @property
    def destination(self):
        """Backup destination info

        """
        rv, stdout, stderr = self.__run__('destinationinfo')
        return TimeMachineDestination(stdout)

    def enable(self):
        """Enable backups

        Requires root
        """
        if os.geteuid() != 0:
            raise TimeMachineError('Error enabling timemachine backups: must be root')

        rv, stdout, stderr = self.__run__('enable')
        if rv != 0:
            raise TimeMachineError('Error enabling backups: {0}{1}'.format(
                ' '.join(stdout.rstrip(), stderr.rstrip()))
            )

    def disable(self):
        """Disable backups

        Requires root
        """
        if os.geteuid() != 0:
            raise TimeMachineError('Error disabling timemachine backups: must be root')

        rv, stdout, stderr = self.__run__('disable')
        if rv != 0:
            raise TimeMachineError('Error disabling backups: {0}{1}'.format(
                ' '.join(stdout.rstrip(), stderr.rstrip()))
            )



"""
Classes to handle AFP mountpoints, mounting and configuring them.
"""

import os
import re

from configobj import ConfigObj
from subprocess import call, CalledProcessError

re_mountpoint = re.compile(r'^/Volumes/[A-Za-z0-9_-]*$')

DEFAULT_CONFIG_PATH = os.path.join(os.getenv('HOME'), '.afpshares.conf')


class AFPShareError(Exception):
    """
    Exceptions raised when accessing AFP shares
    """
    pass


class AFPShareConfig(dict):
    """
    Reader for AFP share configuration files.
    """

    def __init__(self, config_path=DEFAULT_CONFIG_PATH):
        self.config = None
        self.path = config_path

        if not os.path.isfile(self.path):
            raise AFPShareError('No such file: {0}'.format(self.path))

        if not os.access(self.path, os.R_OK):
            raise AFPShareError('No permission to read configuration: {0}'.format(self.path))

        self.config = ConfigObj(self.path)
        if 'Options' in self.config:
            self.options = dict(self.config['Options'])
        else:
            self.options = {}

        for key in [key for key in self.conf if key != 'Options']:
            try:
                self[key] = AFPShareDisk(key, self.config[key])
            except AFPShareError as e:
                raise AFPShareError('Error parsing configuration: {0}'.format(e))

    def __getattr__(self, attr):
        if attr == 'disks':
            return self.values()
        try:
            return self.options[attr]
        except KeyError:
            pass
        raise AttributeError('No such AFPShareConfig attribute: {0}'.format(attr))

    def __getitem__(self, item):
        try:
            return self[item]
        except KeyError:
            pass
        raise KeyError('No such AFP share: {0}'.format(item))


class AFPShareDisk(dict):
    """
    AFP network share disk specification
    """
    def __init__(self, name, settings):
        self.name = name

        self.update(**{
            'address': None,
            'path': None,
            'mountpoint': None,
            'username': None,
            'password': None,
        })

        for key, value in settings.items():
            if key not in self:
                raise AFPShareError('Unsupported disk option: {0}'.format(key))
            self[key] = value

        for key in ('address', 'path', 'mountpoint'):
            if self[key] is None:
                raise AFPShareError('Missing required option {0}'.format(key))

        if 'username' in self and 'password' not in self:
            raise AFPShareError('Username defined but no password given')

        if 'password' in self and 'username' not in self:
            raise AFPShareError('Password defined but no username given')

        if not re.match(re_mountpoint, self.mountpoint):
            raise AFPShareError('Unsupported mountpoint path: {0}'.format(self.mountpoint))

    def __getattr__(self, attr):
        if attr == 'afp_path_nopass':
            return 'afp://%(address)s%(path)s' % self

        if attr == 'afp_path':
            if self.username is not None:
                return 'afp://%(username)s:%(password)s@%(address)s%(path)s' % self
            else:
                return 'afp://%(address)s%(path)s' % self

        try:
            return self[attr]
        except KeyError:
            pass

        raise AttributeError('No such {0} attr: {1}'.format(self.name, attr))

    def __str__(self):
        return '{0}: {1} mounted on {2}'.format(
            self.name,
            self.afp_path_nopass,
            self.mountpoint,
        )

    def status(self):
        """
        Return status of mountpoint as string:

        'not mounted', 'mounted by other user', 'mounted by myself'
        """
        if not os.path.ismount(self.mountpoint):
            return 'not mounted'

        if not os.access(self.mountpoint, os.W_OK):
            return 'mounted by other user'

        else:
            return 'mounted by myself'

    def mount(self):
        """
        Mount the AFP share
        """
        if os.path.ismount(self.mountpoint):
            raise AFPShareError('Already mounted: {0}'.format(self.mountpoint))

        if not os.path.isdir(self.mountpoint):
            try:
                os.makedirs(self.mountpoint)
                if os.stat(self.mountpoint).st_uid != os.getuid():
                    os.chown(self.mountpoint, os.getpid(), os.getgid())

            except IOError as e:
                raise AFPShareError(e)

            except OSError as e:
                raise AFPShareError(e)

        cmd = ('mount_afp', self.afp_path, self.mountpoint)
        try:
            call(cmd)
        except CalledProcessError as e:
            raise AFPShareError('Error running {0}: {1}'.format(' '.join(cmd), e))

    def umount(self):
        """
        Umount the AFP share
        """
        if not os.path.ismount(self.mountpoint):
            return

        if not os.access(self.mountpoint, os.W_OK):
            raise AFPShareError('No write access to {0}'.format(self.mountpoint))

        cmd = ('hdiutil', 'detach', self.mountpoint)
        try:
            call(cmd)
        except CalledProcessError as e:
            raise AFPShareError('Error running {0}: {1}'.format(' '.join(cmd), e))



"""
Module to create application dictionaries for appscript using command line sdef tool
"""

import os
from subprocess import check_output
from lxml import etree as ET


class SDEFError(Exception):
    pass


class SDEF(object):
    def __init__(self, path):
        self.path = path
        if not os.path.isdir(path):
            raise SDEFError('Not a directory: {0}'.format(path))
        self.tree = ET.fromstring(check_output(['sdef', path]))

    def __getattr__(self, attr):
        if attr == 'terms':
            return self.__generate_terms()

    def __generate_terms(self):
        output = 'version = 1.1\npath = {0}'.format(self.path)
        classes = []
        enums = []
        properties = []
        elements = []
        commands = []
        for suite in self.tree.xpath('suite'):
            for node in suite.xpath('class'):
                name = node.get('name').replace(' ', '_')
                code = node.get('code')
                classes.append((name, code))
                if node.get('inherits') is None:
                    continue
                element = node.get('plural').replace('  _')
                elements.append((element, code))

            for node in suite.xpath('enumeration/enumerator'):
                name = node.get('name').replace(' ', '_')
                code = node.get('code')
                enums.append((name, code))

            for node in suite.xpath('class/property'):
                name = node.get('name').replace(' ', '_')
                code = node.get('code')
                properties.append((name, code))

            for node in suite.xpath('command'):
                name = node.get('name').replace(' ', '_')
                code = node.get('code')
                cparams = []
                for p in node.xpath('parameter'):
                    pname = p.get('name').replace(' ', '_')
                    pcode = p.get('code')
                    cparams.append((pname, pcode))
                commands.append((name, code, cparams))

        output += '\nclasses = {0}'.format(classes)
        output += '\nenums = {0}'.format(enums)
        output += '\nproperties = {0}'.format(properties)
        output += '\nelements = {0}'.format(elements)
        output += '\ncommands = {0)}'.format(commands)

        return output




import glob
from setuptools import setup, find_packages
from darwinist import __version__

setup(
    name='darwinist',
    keywords='System Management Utility OS/X Darwin Scripts',
    description='Sysadmin utility modules and scripts for OS/X',
    author='Ilkka Tuohela',
    author_email='hile@iki.fi',
    version=__version__,
    url='https://github.com/hile/darwinist',
    license='PSF',
    scripts=glob.glob('bin/*'),
    packages=find_packages(),
    install_requires=(
        'systematic>=4.8.3',
        'appscript',
    ),
)



"""
Parse and list network interfaces on the system
"""

import re
from subprocess import Popen, PIPE
from seine.address import EthernetMACAddress, IPv4Address, IPv6Address

RE_INET_LINE = [
    re.compile('^\s+inet\s+(?P<address>[0-9.]+)\s+netmask\s+(?P<netmask>[.x0-9a-f]+)$'),
    re.compile('^\s+inet\s+(?P<address>[0-9.]+)\s+netmask\s+(?P<netmask>[.x0-9a-f]+)\s+broadcast\s+(?P<broadcast>[0-9.]+)$'), # noqa
]
RE_INET6_LINE = [
    re.compile('^\s+inet6\s+(?P<address>[0-9:a-f]+)\s+prefixlen\s+(?P<prefix>\d+)$'),
    re.compile('^\s+inet6\s+(?P<address>[0-9:a-f]+)%(?P<scope>[^s]+)\s+prefixlen\s+(?P<prefix>\d+)\s+scopeid\s+(?P<scope_id>[x0-9a-f]+)$'), # noqa
]


class NetworkInterfaces(list):
    def __init__(self):
        p = Popen(['/sbin/ifconfig'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()

        interface = None
        for line in [x.rstrip() for x in stdout.split('\n')]:
            if line.strip() == '':
                continue

            if line.startswith('	'):
                interface.parse(line)

            else:
                name, flags = line.split(':', 1)
                if interface:
                    self.append(interface)
                interface = Interface(name, flags)


class Interface(dict):
    def __init__(self, name, flags=''):
        self.name = name
        self.flags = flags
        self['addresses'] = []

    def parse(self, line):
        """
        Parse a configuration line for interface
        """

        for inet_re in RE_INET_LINE:
            m = inet_re.match(line)
            if m:
                data = m.groupdict()
                data['addr_type'] = 'IPv4'
                data['address'] = IPv4Address(data['address'])
                data['netmask'] = IPv4Address(data['netmask'])
                if 'broadcast' in data:
                    data['broadcast'] = IPv4Address(data['broadcast'])
                else:
                    data['broadcast'] = None

                self['addresses'].append(data)
                return

        for inet6_re in RE_INET6_LINE:
            m = inet6_re.match(line)
            if m:
                data = m.groupdict()
                data['addr_type'] = 'IPv6'
                data['address'] = IPv6Address(data['address'])
                if 'prefix' in data:
                    data['prefix'] = int(data['prefix'])

                self['addresses'].append(data)
                return

        try:
            key, value = line.strip().split(None, 1)
            if key == 'ether':
                value = EthernetMACAddress(value)

        except ValueError:
            try:
                key, value = line.strip().split('=', 1)
            except ValueError:
                raise ValueError('Error splitting line {0}'.format(line))

        self[key] = value



"""
Module for Apple OS/X airport status command access from python
"""

from operator import itemgetter
import os
from subprocess import check_output, CalledProcessError

AIRPORT_BINARY = '/System/Library/PrivateFrameworks/Apple80211.framework/Versions/Current/Resources/airport'


class AirportError(Exception):
    pass


class AirportStatus(dict):
    """
    Class to call the system 'airport' command.
    """
    def __init__(self):
        if not os.path.isfile(AIRPORT_BINARY):
            raise AirportError('No such command: {0}'.format(AIRPORT_BINARY))

    def __repr__(self):
        self.probe()
        return '%(BSSID)s %(SSID)s channel %(channel)s %(agrCtlRSSI)s dB' % self

    def __getattr__(self, attr):
        try:
            if not self.keys():
                self.probe()
            return self[attr]

        except KeyError:
            raise AttributeError('No such AirportStatus attribute: {0}'.format(attr))

    def probe(self):
        """
        Probe airport status
        """
        self.clear()

        cmd = (AIRPORT_BINARY, '-I')
        try:
            data = check_output(cmd)
        except CalledProcessError as e:
            raise AirportError('Error running {0}: {1}'.format(' '.join(cmd), e))

        for line in [line.decode('utf-8') for line in data.splitlines()]:
            if not line.strip():
                continue

            try:
                key, value = [x.strip() for x in line.split(':', 1)]
                self[key] = value
            except ValueError:
                raise AirportError('Error parsing line: {0}'.format(line))

        for key in ('BSSID',):
            if key not in self:
                continue
            self[key] = ':'.join(['%02x'.upper() % int(x, 16) for x in self[key].split(':')])

    def proximity(self):
        """
        Return proximity of base stations based on signal levels
        """
        headers = ('SSID', 'BSSID', 'RSSI', 'CHANNEL', 'HT', 'CC', 'SECURITY')
        aps = []

        cmd = (AIRPORT_BINARY, '-s', self.SSID)
        try:
            data = check_output(cmd)
        except CalledProcessError as e:
            raise AirportError('Error running {0}: {1}'.format(' '.join(cmd), e))
            return

        for line in [line.decode('utf-8') for line in data.splitlines()]:
            if not line.strip():
                continue

            line = line.rstrip()
            if headers[:5] == [x.strip() for x in line.split()][:5]:
                continue

            ssid = line[:32].lstrip()
            bssid = line[33:50].strip("'").upper()
            rssi = int(line[51:55].strip())
            channel = int(line[56:58].strip())
            aps.append({
                'SSID': ssid,
                'BSSID': bssid,
                'RSSI': rssi,
                'CHANNEL': channel,
            })

        aps.sort(key=itemgetter('RSSI'))
        return aps
